@@ -22,6 +22,7 @@
 #include "Dir.h"
 #include "Language.h"
 #include "Resource.h"
+#include "LzmaLib.h"
 #include <tchar.h>
 #include <Strsafe.h>
 
@@ -32,8 +33,12 @@
 #ifdef PORTABLE
 #define OutputPackageFile L"VeraCrypt Portable " _T(VERSION_STRING) _T(VERSION_STRING_SUFFIX)L".exe"
 #else
+#ifdef VC_COMREG
+#define OutputPackageFile L"VeraCrypt COMReg.exe"
+#else
 #define OutputPackageFile L"VeraCrypt Setup " _T(VERSION_STRING) _T(VERSION_STRING_SUFFIX) L".exe"
 #endif
+#endif
 #define MAG_START_MARKER	"VCINSTRT"
 #define MAG_END_MARKER_OBFUSCATED	"V/C/I/N/S/C/R/C"
 #define PIPE_BUFFER_LEN	(4 * BYTES_PER_KB)
@@ -41,6 +46,7 @@
 unsigned char MagEndMarker [sizeof (MAG_END_MARKER_OBFUSCATED)];
 wchar_t DestExtractPath [TC_MAX_PATH];
 DECOMPRESSED_FILE	Decompressed_Files [NBR_COMPRESSED_FILES];
+int Decompressed_Files_Count = 0;
 
 volatile char *PipeWriteBuf = NULL;
 volatile HANDLE hChildStdinWrite = INVALID_HANDLE_VALUE;
@@ -89,24 +95,43 @@ static void PkgInfo (wchar_t *msg)
 // Returns 0 if decompression fails or, if successful, returns the size of the decompressed data
 static int DecompressBuffer (unsigned char *out, int outSize, unsigned char *in, int len)
 {
-	uLongf outlen = (uLongf) outSize;
-	int ret = uncompress (out, &outlen, in, (uLong) len);
-	if (Z_OK == ret)
-		return (int) outlen;
-	else
-		return 0;
+	int outlen = 0;
+	int status;
+	if (len > 5)
+	{
+		// the first 5 bytes of in contain props parameter
+		size_t srcLen = len - 5;
+		size_t outputLen = (size_t) outSize;
+		status = LzmaUncompress (out, &outputLen, in + 5, &srcLen, in, 5);
+		if (status == SZ_OK)
+		{
+			outlen = (int) outputLen;
+		}
+	}
+
+	return outlen;
 }
 
 
 // Returns 0 if compression fails or, if successful, the size of the compressed data
 static int CompressBuffer (unsigned char *out, int outSize, unsigned char *in, int len)
 {
-	uLongf outlen = (uLongf) outSize;
-	int ret = compress2 (out, &outlen, in, (uLong) len, Z_BEST_COMPRESSION);
-	if (Z_OK == ret)
-		return (int) outlen;
-	else
-		return 0;
+	unsigned char outProps[5];
+	size_t outPropsSize = 5;
+	int outlen = 0;
+	int status;
+	if (outSize > 5)
+	{
+		size_t outputLen = (size_t) (outSize - 5);
+		status = LzmaCompress (out + 5, &outputLen, in, len, outProps, &outPropsSize, 9, 0, -1, -1, -1, -1, -1);
+		if (status == SZ_OK)
+		{
+			memcpy (out, outProps, outPropsSize);
+			outlen = (int) (outputLen + 5);
+		}
+	}
+
+	return outlen;
 }
 
 
@@ -119,7 +144,7 @@ static void WipeSignatureAreas (char *buffer)
 }
 
 
-BOOL MakeSelfExtractingPackage (HWND hwndDlg, wchar_t *szDestDir)
+BOOL MakeSelfExtractingPackage (HWND hwndDlg, wchar_t *szDestDir, BOOL bSkipX64)
 {
 	int i, x;
 	wchar_t inputFile [TC_MAX_PATH];
@@ -163,6 +188,17 @@ BOOL MakeSelfExtractingPackage (HWND hwndDlg, wchar_t *szDestDir)
 
 	for (i = 0; i < sizeof (szCompressedFiles) / sizeof (szCompressedFiles[0]); i++)
 	{
+		if (bSkipX64 && wcsstr(szCompressedFiles[i], L"-x64"))
+			continue;
+
+#ifdef VC_COMREG
+		if (	wcsstr(szCompressedFiles[i], L".zip") || wcsstr(szCompressedFiles[i], L".inf") 
+			||	wcsstr(szCompressedFiles[i], L".cat") || wcsstr(szCompressedFiles[i], L".txt")
+			||	wcsstr(szCompressedFiles[i], L"LICENSE") || wcsstr(szCompressedFiles[i], L"NOTICE")
+			)
+			continue;
+#endif
+
 		StringCbPrintfW (szTmpFilePath, sizeof(szTmpFilePath), L"%s%s", szDestDir, szCompressedFiles[i]);
 
 		if (!FileExists (szTmpFilePath))
@@ -215,6 +251,17 @@ BOOL MakeSelfExtractingPackage (HWND hwndDlg, wchar_t *szDestDir)
 		DWORD tmpFileSize;
 		unsigned char *tmpBuffer;
 
+		if (bSkipX64 && wcsstr(szCompressedFiles[i], L"-x64"))
+			continue;
+
+#ifdef VC_COMREG
+		if (	wcsstr(szCompressedFiles[i], L".zip") || wcsstr(szCompressedFiles[i], L".inf") 
+			||	wcsstr(szCompressedFiles[i], L".cat") || wcsstr(szCompressedFiles[i], L".txt")
+			||	wcsstr(szCompressedFiles[i], L"LICENSE") || wcsstr(szCompressedFiles[i], L"NOTICE")
+			)
+			continue;
+#endif
+
 		StringCbPrintfW (szTmpFilePath, sizeof(szTmpFilePath), L"%s%s", szDestDir, szCompressedFiles[i]);
 
 		tmpBuffer = LoadFile (szTmpFilePath, &tmpFileSize);
@@ -485,6 +532,8 @@ void FreeAllFileBuffers (void)
 		Decompressed_Files[fileNo].fileLength = 0;
 		Decompressed_Files[fileNo].crc = 0;
 	}
+
+	Decompressed_Files_Count = 0;
 }
 
 
@@ -492,7 +541,7 @@ void FreeAllFileBuffers (void)
 // Creates a table of pointers to buffers containing the following objects for each file:
 // filename size, filename (not null-terminated!), file size, file CRC-32, uncompressed file contents.
 // For details, see the definition of the DECOMPRESSED_FILE structure.
-BOOL SelfExtractInMemory (wchar_t *path)
+BOOL SelfExtractInMemory (wchar_t *path, BOOL bSkipCountCheck)
 {
 	int filePos = 0, fileNo = 0;
 	int fileDataEndPos = 0;
@@ -548,7 +597,7 @@ BOOL SelfExtractInMemory (wchar_t *path)
 		Error ("DIST_PACKAGE_CORRUPTED", NULL);
 	}
 
-	decompressedDataLen = uncompressedLen + 524288;	// + 512K reserve
+	decompressedDataLen = uncompressedLen;
 	DecompressedData = malloc (decompressedDataLen);
 	if (DecompressedData == NULL)
 	{
@@ -607,12 +656,14 @@ BOOL SelfExtractInMemory (wchar_t *path)
 		fileNo++;
 	}
 
-	if (fileNo < NBR_COMPRESSED_FILES)
+	if (!bSkipCountCheck && (fileNo < NBR_COMPRESSED_FILES))
 	{
 		Error ("DIST_PACKAGE_CORRUPTED", NULL);
 		goto sem_end;
 	}
 
+	Decompressed_Files_Count = fileNo;
+
 	free (compressedData);
 	return TRUE;
 
@@ -635,7 +686,7 @@ void __cdecl ExtractAllFilesThread (void *hwndDlg)
 
 	GetModuleFileName (NULL, packageFile, ARRAYSIZE (packageFile));
 
-	if (!(bSuccess = SelfExtractInMemory (packageFile)))
+	if (!(bSuccess = SelfExtractInMemory (packageFile, FALSE)))
 		goto eaf_end;
 
 	if (mkfulldir (DestExtractPath, TRUE) != 0)
